import random
import pickle
import CADETMatch.util as util
import numpy
import numpy as np
import scipy
from pathlib import Path
import time
import csv
import cadet

import emcee
import SALib.sample.sobol_sequence

import subprocess
import sys

import CADETMatch.evo as evo
import CADETMatch.cache as cache

import CADETMatch.pareto as pareto

import multiprocessing
import pandas
import array
import emcee.autocorr as autocorr

import CADETMatch.kde_generator as kde_generator
from sklearn.neighbors.kde import KernelDensity

name = "MCMC"

from addict import Dict

import joblib

import CADETMatch.de as de
import CADETMatch.de_snooker as de_snooker
import CADETMatch.stretch as stretch

import json
import shutil

log2 = numpy.log(2)

acceptance_target = 0.234
acceptance_delta = 0.1

def log_previous(cadetValues, kde_previous, kde_previous_scaler):
    #find the right values to use
    col = len(kde_previous_scaler.scale_)
    values = cadetValues[-col:]
    values_shape = numpy.array(values).reshape(1, -1)
    values_scaler = kde_previous_scaler.transform(values_shape)
    score = kde_previous.score_samples(values_scaler)
    return score

def log_likelihood(individual, json_path):
    if json_path != cache.cache.json_path:
        cache.cache.setup_dir(json_path)
        util.setupLog(cache.cache.settings['resultsDirLog'], "main.log")
        cache.cache.setup(json_path, False)

    kde_previous, kde_previous_scaler = kde_generator.getKDEPrevious(cache.cache)

    if 'kde' not in log_likelihood.__dict__:
        kde, kde_scaler = kde_generator.getKDE(cache.cache)
        log_likelihood.kde = kde
        log_likelihood.scaler = kde_scaler

    scores, csv_record, results, individual = evo.fitness(individual, json_path)

    if results is None:
        return -numpy.inf, scores, csv_record, results, individual

    if results is not None and kde_previous is not None:
        logPrevious = log_previous(individual, kde_previous, kde_previous_scaler)
    else:
        logPrevious = 0.0

    scores_shape = numpy.array(scores).reshape(1, -1)

    score_scaler = log_likelihood.scaler.transform(scores_shape)

    score_kde = log_likelihood.kde.score_samples(score_scaler)

    score = score_kde + log2 + logPrevious #*2 is from mirroring and we need to double the probability to get back to the normalized distribution

    return score, scores, csv_record, results, individual 

def log_posterior_vectorize(population, json_path, cache, halloffame, meta_hof, grad_hof, result_data, writer, csvfile):
    results = cache.toolbox.map(log_posterior, ( (population[i], json_path) for i in range(len(population))))
    results = process(population, cache, halloffame, meta_hof, grad_hof, result_data, results, writer, csvfile)
    return results

def log_posterior(x):
    theta, json_path = x
    if json_path != cache.cache.json_path:
        cache.cache.setup_dir(json_path)
        util.setupLog(cache.cache.settings['resultsDirLog'], "main.log")
        cache.cache.setup(json_path)

    ll, scores, csv_record, results, individual = log_likelihood(theta, json_path)
    if results is None:
        return -numpy.inf, None, None, None, None, individual
    else:
        return ll, theta, scores, csv_record, results, individual

def addChain(*args):
    temp = [arg for arg in args if arg is not None]
    if len(temp) > 1:
        return numpy.concatenate( temp, axis=1)
    else:
        return numpy.array(temp[0])

def converged_bounds(chain, length, error_level):
    if chain.shape[1] < (2*length):
        return False, None, None
    lb = []
    ub = []
    
    start = chain.shape[1] - length
    stop = chain.shape[1]
    for i in range(start, stop):
        temp_chain = chain[:,:i,:]
        temp_chain_shape = temp_chain.shape
        temp_chain_flat = temp_chain.reshape(temp_chain_shape[0] * temp_chain_shape[1], temp_chain_shape[2])
        lb_5, ub_95 = numpy.percentile(temp_chain_flat, [5, 95], 0)
        
        lb.append(lb_5)
        ub.append(ub_95)
       
    lb = numpy.array(lb)
    ub = numpy.array(ub)
    
    if numpy.all(numpy.std(lb, axis=0) < error_level) and numpy.all(numpy.std(ub, axis=0) < error_level):
        return True, numpy.mean(lb, axis=0), numpy.mean(ub, axis=0)
    else:
        multiprocessing.get_logger().info("bounds have not yet converged lb min: %s max: %s std: %s  ub min %s max: %s std: %s", 
                                          numpy.array2string(numpy.min(lb, axis=0), precision=3, separator=','),  
                                          numpy.array2string(numpy.max(lb, axis=0), precision=3, separator=','), 
                                          numpy.array2string(numpy.std(lb, axis=0), precision=4, separator=','),
                                          numpy.array2string(numpy.min(ub, axis=0), precision=3, separator=','),  
                                          numpy.array2string(numpy.max(ub, axis=0), precision=3, separator=','), 
                                          numpy.array2string(numpy.std(ub, axis=0), precision=4, separator=','), )
        return False, None, None

def rescale(lb, ub, old_lb, old_ub):
    "give a new lb and ub that will rescale so that the previous lb and ub takes up about 1/2 of the search width"
    center = (ub+lb)/2.0
    
    new_lb = lb - 2*(center - lb)

    new_lb = numpy.max([new_lb, old_lb],axis=0)
    
    new_ub = ub + 2*(ub - center)

    new_ub = numpy.min([new_ub, old_ub],axis=0)
    
    return new_lb, center, new_ub

def flatten(chain):
    chain_shape = chain.shape
    flat_chain = chain.reshape(chain_shape[0] * chain_shape[1], chain_shape[2])
    return flat_chain

def change_bounds_json(cache, lb, ub):
    "change the bounds based on lb and ub and then save it as a new json file and return the path to the new file"
    multiprocessing.get_logger().info("change_bounds_json  lb %s  ub %s", lb, ub)
    lb_trans = util.convert_individual(lb, cache)[1]
    ub_trans = util.convert_individual(ub, cache)[1]

    settings_file = Path(cache.json_path)
    settings_file_backup = settings_file.with_suffix('.json.backup')

    new_name = "%s_bounds%s" % (settings_file.stem, settings_file.suffix)

    new_settings_file = settings_file.with_name(new_name)

    with settings_file.open() as json_data:
        settings = json.load(json_data)

        idx = 0
        for parameter in settings['parameters']:
            transform = cache.transforms[parameter['transform']]
            count = transform.count_extended
            multiprocessing.get_logger().warn('%s %s %s', idx, count, transform)
            lb_local = lb_trans[idx:idx+count]
            ub_local = ub_trans[idx:idx+count]
            transform.setBounds(parameter, lb_local, ub_local)
            idx = idx + count

        with new_settings_file.open(mode="w") as json_data:
            json.dump(settings, json_data, indent=4, sort_keys=False)

    #copy the original file to a backup name
    shutil.copy(settings_file, settings_file_backup)

    #copy over our new settings file to the original file also
    #this is so that external programs also see the new bounds
    with settings_file.open(mode="w") as json_data:
        json.dump(settings, json_data, indent=4, sort_keys=False)

    return new_settings_file.as_posix()

def sampler_auto_bounds(cache, checkpoint, sampler, checkpointFile):
    burn_seq = checkpoint.get('burn_seq', [])
    chain_seq = checkpoint.get('chain_seq', [])
    bounds_seq = checkpoint.get('bounds_seq', [])
        
    train_chain = checkpoint.get('train_chain', None)
    run_chain = checkpoint.get('run_chain', None)
    bounds_chain = checkpoint.get('bounds_chain', None)

    train_chain_stat = checkpoint.get('train_chain_stat', None)
    run_chain_stat = checkpoint.get('run_chain_stat', None)

    checkInterval = 25

    parameters = len(cache.MIN_VALUE)

    new_parameters = len(cache.settings['parameters']) - len(cache.settings.get('parameters_mcmc', []))

    finished = False

    generation = checkpoint['idx_bounds']

    sampler.iterations = checkpoint['sampler_iterations']
    sampler.naccepted = checkpoint['sampler_naccepted']
    sampler._moves[1].n = checkpoint['sampler_n']
    tau_percent = None

    while not finished:
        state = next(sampler.sample(checkpoint['p_bounds'], log_prob0=checkpoint['ln_prob_bounds'], rstate0=checkpoint['rstate_bounds'], iterations=1 ))

        p = state.coords
        ln_prob = state.log_prob
        random_state = state.random_state

        accept = numpy.mean(sampler.acceptance_fraction)
        bounds_seq.append(accept)

        bounds_chain = addChain(bounds_chain, p[:, numpy.newaxis, :])

        multiprocessing.get_logger().info('run:  idx: %s accept: %.3f', generation, accept)
        
        generation += 1

        checkpoint['p_bounds'] = p
        checkpoint['ln_prob_bounds'] = ln_prob
        checkpoint['rstate_bounds'] = random_state
        checkpoint['idx_bounds'] = generation
        checkpoint['bounds_chain'] = bounds_chain
        checkpoint['bounds_seq'] = bounds_seq
        checkpoint['bounds_iterations'] = sampler.iterations
        checkpoint['bounds_naccepted'] = sampler.naccepted

        write_interval(cache.checkpointInterval, cache, checkpoint, checkpointFile, bounds_chain, train_chain, run_chain, bounds_seq, burn_seq, chain_seq, parameters, train_chain_stat, run_chain_stat, tau_percent)
        util.graph_corner_process(cache, last=False)

        if generation % checkInterval == 0:
            converged, lb, ub = converged_bounds(bounds_chain[:,:,:new_parameters], 200, 1e-3)

            if converged:
                finished = True

                new_lb, center, new_ub = rescale(lb, ub, numpy.array(cache.MIN_VALUE[:new_parameters]), numpy.array(cache.MAX_VALUE[:new_parameters]))

                new_min_value = list(cache.MIN_VALUE)
                new_max_value = list(cache.MAX_VALUE)

                new_min_value[:new_parameters] = new_lb
                new_max_value[:new_parameters] = new_ub
                json_path = change_bounds_json(cache, new_min_value, new_max_value)                
                cache.resetTransform(json_path)                
                sampler.log_prob_fn.args[0] = json_path
                sampler.log_prob_fn.args[1] = cache

                multiprocessing.get_logger().info("new lower bounds %s   new upper bounds %s", new_lb, new_ub)
            else:
                multiprocessing.get_logger().info("bounds have not yet converged in gen %s", generation)

    sampler.reset()
    checkpoint['state'] = 'burn_in'
    checkpoint['p_burn'] = p

    write_interval(-1, cache, checkpoint, checkpointFile, bounds_chain, train_chain, run_chain, bounds_seq, burn_seq, chain_seq, parameters, train_chain_stat, run_chain_stat, tau_percent)

def sampler_burn(cache, checkpoint, sampler, checkpointFile):
    burn_seq = checkpoint.get('burn_seq', [])
    chain_seq = checkpoint.get('chain_seq', [])
    bounds_seq = checkpoint.get('bounds_seq', [])
        
    train_chain = checkpoint.get('train_chain', None)
    run_chain = checkpoint.get('run_chain', None)
    bounds_chain = checkpoint.get('bounds_chain', None)

    train_chain_stat = checkpoint.get('train_chain_stat', None)
    run_chain_stat = checkpoint.get('run_chain_stat', None)

    converge = checkpoint.get('converge')    

    parameters = len(cache.MIN_VALUE)

    tol = 5e-4
    power = 0.0
    distance = 1.0
    #distance_a = sampler.a
    stop_next = False
    finished = False

    generation = checkpoint['idx_burn']

    sampler.iterations = checkpoint['sampler_iterations']
    sampler.naccepted = checkpoint['sampler_naccepted']
    sampler._moves[1].n = checkpoint['sampler_n']

    while not finished:
        state = next(sampler.sample(checkpoint['p_burn'], log_prob0=checkpoint['ln_prob_burn'], rstate0=checkpoint['rstate_burn'], iterations=1,
                                                        tune=False))

        p = state.coords
        ln_prob = state.log_prob
        random_state = state.random_state

        accept = numpy.mean(sampler.acceptance_fraction)
        burn_seq.append(accept)
        converge[:-1] = converge[1:]
        converge[-1] = accept

        train_chain = addChain(train_chain, p[:, numpy.newaxis, :])

        train_chain_stat = addChain(train_chain_stat, numpy.percentile(flatten(train_chain), [5, 50, 95], 0)[:, numpy.newaxis, :])

        converge_real = converge[~numpy.isnan(converge)]
        multiprocessing.get_logger().info('burn:  idx: %s accept: %.3g std: %.3g mean: %.3g converge: %.3g', generation, accept, 
                            numpy.std(converge_real), numpy.mean(converge_real), numpy.std(converge_real)/tol)

        generation += 1

        checkpoint['p_burn'] = p
        checkpoint['ln_prob_burn'] = ln_prob
        checkpoint['rstate_burn'] = random_state
        checkpoint['idx_burn'] = generation
        checkpoint['train_chain'] = train_chain
        checkpoint['burn_seq'] = burn_seq
        checkpoint['converge'] = converge
        checkpoint['sampler_iterations'] = sampler.iterations
        checkpoint['sampler_naccepted'] = sampler.naccepted
        checkpoint['train_chain_stat'] = train_chain_stat
        checkpoint['run_chain_stat'] = run_chain_stat
        checkpoint['sampler_n'] = sampler._moves[1].n

        write_interval(cache.checkpointInterval, cache, checkpoint, checkpointFile, bounds_chain, train_chain, run_chain, bounds_seq, burn_seq, chain_seq, parameters, train_chain_stat, run_chain_stat, None)
        util.graph_corner_process(cache, last=False)

        if numpy.std(converge_real) < tol and len(converge) == len(converge_real):
            average_converge = numpy.mean(converge)
            if (acceptance_target - acceptance_delta) < average_converge < (acceptance_target + acceptance_delta):
                multiprocessing.get_logger().info("burn in completed at iteration %s", generation)
                finished = True

            if stop_next is True:
                multiprocessing.get_logger().info("burn in completed at iteration %s based on minimum distances", generation)
                finished = True

            if not finished:
                new_distance = numpy.abs(average_converge - acceptance_target)
                if new_distance < distance:
                    distance_n = sampler._moves[1].n
                    distance = new_distance

                    multiprocessing.get_logger().info("burn in acceptance is out of tolerance and n must be adjusted while burn in continues")
                    converge[:] = numpy.nan
                    prev_n = sampler._moves[1].n
                    if average_converge > (acceptance_target + 3 * acceptance_delta):
                        #n must be increased to decrease the acceptance rate (step size)
                        power += 4
                    elif average_converge > (acceptance_target + 2 * acceptance_delta):
                        #n must be increased to decrease the acceptance rate (step size)
                        power += 2
                    elif average_converge > (acceptance_target + 1 * acceptance_delta):
                        #n must be increased to decrease the acceptance rate (step size)
                        power += 1                    
                    elif average_converge < (acceptance_target - 3 * acceptance_delta):
                        #n must be decreased to increase the acceptance rate (step size)
                        power -= 4
                    elif average_converge < (acceptance_target - 2 * acceptance_delta):
                        #n must be decreased to increase the acceptance rate (step size)
                        power -= 2
                    elif average_converge < (acceptance_target - 1 * acceptance_delta):
                        #n must be decreased to increase the acceptance rate (step size)
                        power -= 1
                    new_n = power
                    sampler._moves[1].n = power
                    sampler.reset()
                    checkpoint['p_burn'] = checkpoint['starting_population']
                    checkpoint['ln_prob_burn'] = None
                    multiprocessing.get_logger().info('previous n: %s    new n: %s', prev_n, new_n)
                else:
                    sampler._moves[1].n = distance_n
                    sampler.reset()
                    checkpoint['p_burn'] = checkpoint['starting_population']
                    checkpoint['ln_prob_burn'] = None
                    stop_next = True
 
    sampler.reset()

    checkpoint['sampler_iterations'] = sampler.iterations
    checkpoint['sampler_naccepted'] = sampler.naccepted
    checkpoint['state'] = 'chain'
    checkpoint['p_chain'] = p
    checkpoint['ln_prob_chain'] = None
    checkpoint['rstate_chain'] = None
    checkpoint['sampler_a'] = sampler._moves[1].n

    write_interval(-1, cache, checkpoint, checkpointFile, bounds_chain, train_chain, run_chain, bounds_seq, burn_seq, chain_seq, parameters, train_chain_stat, run_chain_stat, None)
            

def sampler_run(cache, checkpoint, sampler, checkpointFile):
    burn_seq = checkpoint.get('burn_seq', [])
    chain_seq = checkpoint.get('chain_seq', [])
    bounds_seq = checkpoint.get('bounds_seq', [])
        
    train_chain = checkpoint.get('train_chain', None)
    run_chain = checkpoint.get('run_chain', None)
    bounds_chain = checkpoint.get('bounds_chain', None)

    train_chain_stat = checkpoint.get('train_chain_stat', None)
    run_chain_stat = checkpoint.get('run_chain_stat', None)

    iat = checkpoint.get('integrated_autocorrelation_time', [])

    checkInterval = 25

    parameters = len(cache.MIN_VALUE)

    finished = False

    generation = checkpoint['idx_chain']

    sampler.iterations = checkpoint['sampler_iterations']
    sampler.naccepted = checkpoint['sampler_naccepted']
    sampler._moves[1].n = checkpoint['sampler_n']
    tau_percent = None

    while not finished:
        state = next(sampler.sample(checkpoint['p_chain'], log_prob0=checkpoint['ln_prob_chain'], rstate0=checkpoint['rstate_chain'], iterations=1 ))

        p = state.coords
        ln_prob = state.log_prob
        random_state = state.random_state

        accept = numpy.mean(sampler.acceptance_fraction)
        chain_seq.append(accept)

        run_chain = addChain(run_chain, p[:, numpy.newaxis, :])

        run_chain_stat = addChain(run_chain_stat, numpy.percentile(flatten(run_chain), [5, 50, 95], 0)[:, numpy.newaxis, :])

        multiprocessing.get_logger().info('run:  idx: %s accept: %.3f', generation, accept)
        
        generation += 1

        checkpoint['p_chain'] = p
        checkpoint['ln_prob_chain'] = ln_prob
        checkpoint['rstate_chain'] = random_state
        checkpoint['idx_chain'] = generation
        checkpoint['run_chain'] = run_chain
        checkpoint['chain_seq'] = chain_seq
        checkpoint['sampler_iterations'] = sampler.iterations
        checkpoint['sampler_naccepted'] = sampler.naccepted
        checkpoint['train_chain_stat'] = train_chain_stat
        checkpoint['run_chain_stat'] = run_chain_stat

        if generation % checkInterval == 0:
            try:
                tau = autocorr.integrated_time(numpy.swapaxes(run_chain, 0, 1), tol=cache.MCMCTauMult)
                multiprocessing.get_logger().info("Mean acceptance fraction: %s %0.3f tau: %s with shape: %s", generation, accept, tau, run_chain.shape)
                if numpy.any(numpy.isnan(tau)):
                    multiprocessing.get_logger().info("tau is NaN and clearly not complete %s", generation)
                else:
                    multiprocessing.get_logger().info("we have run long enough and can quit %s", generation)
                    finished = True
            except autocorr.AutocorrError as err:
                multiprocessing.get_logger().info(str(err))
                tau = err.tau
            multiprocessing.get_logger().info("Mean acceptance fraction: %s %0.3f tau: %s", generation, accept, tau)

            temp_iat = [generation]
            temp_iat.extend(tau)
            iat.append(temp_iat)
            checkpoint['integrated_autocorrelation_time'] = iat

            tau = numpy.array(tau)
            tau_percent = generation / (tau * cache.MCMCTauMult)

        write_interval(cache.checkpointInterval, cache, checkpoint, checkpointFile, bounds_chain, train_chain, run_chain, bounds_seq, burn_seq, chain_seq, parameters, train_chain_stat, run_chain_stat, tau_percent, iat)
        mle_process(last=False)
        util.graph_corner_process(cache, last=False)

    checkpoint['p_chain'] = p
    checkpoint['ln_prob_chain'] = ln_prob
    checkpoint['rstate_chain'] = random_state
    checkpoint['idx_chain'] = generation
    checkpoint['run_chain'] = run_chain
    checkpoint['chain_seq'] = chain_seq
    checkpoint['state'] = 'complete'

    write_interval(-1, cache, checkpoint, checkpointFile, bounds_chain, train_chain, run_chain, bounds_seq, burn_seq, chain_seq, parameters, train_chain_stat, run_chain_stat, tau_percent, iat)

def write_interval(interval, cache, checkpoint, checkpointFile, bounds_chain, train_chain, run_chain, bounds_seq, burn_seq, chain_seq, parameters, train_chain_stat, run_chain_stat, tau_percent=None, iat=None):
    "write the checkpoint and mcmc data at most every n seconds"
    if 'last_time' not in write_interval.__dict__:
        write_interval.last_time = time.time()

    if time.time() - write_interval.last_time > interval:
        with checkpointFile.open('wb') as cp_file:
            pickle.dump(checkpoint, cp_file)

        writeMCMC(cache, bounds_chain, train_chain, run_chain, bounds_seq, burn_seq, chain_seq, parameters, train_chain_stat, run_chain_stat, tau_percent, iat)
        
        write_interval.last_time = time.time()

def run(cache, tools, creator):
    "run the parameter estimation"
    random.seed()
    checkpointFile = Path(cache.settings['resultsDirMisc'], cache.settings.get('checkpointFile', 'check'))
    checkpoint = getCheckPoint(checkpointFile,cache)

    burn_seq = checkpoint.get('burn_seq', [])
    chain_seq = checkpoint.get('chain_seq', [])
        
    train_chain = checkpoint.get('train_chain', None)
    run_chain = checkpoint.get('run_chain', None)

    parameters = len(cache.MIN_VALUE)
    
    MCMCpopulationSet = cache.settings.get('MCMCpopulationSet', None)
    if MCMCpopulationSet is not None:
        populationSize = MCMCpopulationSet
    else:
        populationSize = parameters * cache.settings['MCMCpopulation']
               
    #Population must be even
    populationSize = populationSize + populationSize % 2  

    #due to emcee 3.0 and RedBlueMove there is a minimum population size to work correctly based on the number of paramters
    populationSize = max(parameters*2, populationSize)

    if checkpoint['state'] == 'start':
        multiprocessing.get_logger().info("setting up kde")
        kde, kde_scaler = kde_generator.setupKDE(cache)
        checkpoint['state'] = 'auto_bounds'

        with checkpointFile.open('wb') as cp_file:
            pickle.dump(checkpoint, cp_file)
    else:
        multiprocessing.get_logger().info("loading kde")
        kde, kde_scaler = kde_generator.getKDE(cache)

    path = Path(cache.settings['resultsDirBase'], cache.settings['csv'])
    with path.open('a', newline='') as csvfile:
        writer = csv.writer(csvfile, delimiter=',', quoting=csv.QUOTE_ALL)

        result_data = {'input':[], 'output':[], 'output_meta':[], 'results':{}, 'times':{}, 'input_transform':[], 'input_transform_extended':[], 'strategy':[], 
                   'mean':[], 'confidence':[], 'mcmc_score':[]}
        halloffame = pareto.DummyFront(similar=util.similar)
        meta_hof = pareto.ParetoFront(similar=util.similar, similar_fit=util.similar_fit_meta)
        grad_hof = pareto.DummyFront(similar=util.similar)
                

        sampler = emcee.EnsembleSampler(populationSize, parameters, log_posterior_vectorize, 
                                        args=[cache.json_path, cache,
                                                halloffame, meta_hof, grad_hof, result_data,
                                                writer, csvfile], 
                                        moves=[(de_snooker.DESnookerMove(), 0.1), 
                                               (de.DEMove(), 0.9 * 0.9),
                                               (de.DEMove(gamma0=1.0), 0.9 * 0.1),],
                                        vectorize=True)

        if 'sampler_n' not in checkpoint:
            checkpoint['sampler_n'] = sampler._moves[1].n

        if checkpoint['state'] == 'auto_bounds':
            sampler_auto_bounds(cache, checkpoint, sampler, checkpointFile)

        if checkpoint['state'] == 'burn_in':
            sampler_burn(cache, checkpoint, sampler, checkpointFile)

        if checkpoint['state'] == 'chain':
            run_chain = checkpoint.get('run_chain', None)
            if run_chain is not None:
                temp = run_chain[:, :checkpoint['idx_chain'], 0].T
                multiprocessing.get_logger().info('complete shape %s', temp.shape)
 
                try:
                    tau = autocorr.integrated_time(numpy.swapaxes(run_chain[:, :checkpoint['idx_chain'], :], 0, 1), tol=cache.MCMCTauMult)
                    multiprocessing.get_logger().info("we have previously run long enough and can quit %s", checkpoint['idx_chain'])
                    checkpoint['state'] = 'complete'
                except autocorr.AutocorrError as err:
                    multiprocessing.get_logger().info(str(err))
                    tau = err.tau

                multiprocessing.get_logger().info("Mean acceptance fraction: %s %0.3f tau: %s", checkpoint['idx_chain'], checkpoint['chain_seq'][-1], tau)
            
        if checkpoint['state'] == 'chain':
            sampler_run(cache, checkpoint, sampler, checkpointFile)

    chain = checkpoint['run_chain']
    chain_shape = chain.shape
    chain = chain.reshape(chain_shape[0] * chain_shape[1], chain_shape[2])

    if checkpoint['state'] == 'complete':
        tube_process(last=True)
        util.finish(cache)
        checkpoint['state'] = 'plot_finish'

        with checkpointFile.open('wb') as cp_file:
            pickle.dump(checkpoint, cp_file)

    if checkpoint['state'] == 'plot_finish':
        mle_process(last=True)
        util.graph_corner_process(cache, last=True)
    return numpy.mean(chain, 0)

def tube_process(last=False, interval=3600):
    cwd = str(Path(__file__).parent.parent)
    ret = subprocess.run([sys.executable, 'mcmc_plot_tube.py', str(cache.cache.json_path), str(util.getCoreCounts())], 
            stdin=None, stdout=None, stderr=None, close_fds=True,  cwd=cwd)

def mle_process(last=False, interval=3600):
    if 'last_time' not in mle_process.__dict__:
        mle_process.last_time = time.time()

    if 'child' in mle_process.__dict__:
        if mle_process.child.poll() is None:  #This is false if the child has completed
            if last:
                mle_process.child.wait()
            else:
                return

    cwd = str(Path(__file__).parent.parent)

    if last:
        ret = subprocess.run([sys.executable, 'mle.py', str(cache.cache.json_path), str(util.getCoreCounts())], 
            stdin=None, stdout=None, stderr=None, close_fds=True,  cwd=cwd)
        mle_process.last_time = time.time()
    elif (time.time() - mle_process.last_time) > interval:
        #mle_process.child = subprocess.Popen([sys.executable, 'mle.py', str(cache.cache.json_path), str(util.getCoreCounts())], 
        #    stdin=None, stdout=None, stderr=None, close_fds=True,  cwd=cwd)
        subprocess.run([sys.executable, 'mle.py', str(cache.cache.json_path), str(util.getCoreCounts())], 
            stdin=None, stdout=None, stderr=None, close_fds=True,  cwd=cwd)
        mle_process.last_time = time.time()
        
def get_population(base, size, diff=0.05):
    new_population = base
    row, col = base.shape
    multiprocessing.get_logger().info('%s', base)
    
    multiprocessing.get_logger().info('row %s size %s', row, size)
    if row < size:
        #create new entries
        indexes = numpy.random.choice(new_population.shape[0], size - row, replace=True)
        temp = new_population[indexes,:]
        rand = numpy.random.uniform(1.0-diff, 1.0+diff, size=temp.shape)
        new_population = numpy.concatenate([new_population, temp * rand])
    if row > size:
        #randomly select entries to keep
        indexes = numpy.random.choice(new_population.shape[0], size, replace=False)
        multiprocessing.get_logger().info('indexes: %s', indexes)
        new_population = new_population[indexes,:]
    change = numpy.random.normal(1.0, 0.01, new_population.shape)
    multiprocessing.get_logger().info("Initial population condition number before %s  after %s", numpy.linalg.cond(new_population), numpy.linalg.cond(new_population*change))
    return new_population * change

def resetPopulation(checkpoint, cache):
    populationSize = checkpoint['populationSize']
    parameters = len(cache.MIN_VALUE)

    if cache.settings.get('PreviousResults', None) is not None:
        multiprocessing.get_logger().info('running with previous best results')
        previousResultsFile = Path(cache.settings['PreviousResults'])
        results_h5 = cadet.H5()
        results_h5.filename = previousResultsFile.as_posix()
        results_h5.load()
        previousResults = results_h5.root.meta_population_transform

        row,col = previousResults.shape
        multiprocessing.get_logger().info('row: %s col: %s  parameters: %s', row, col, parameters)
        if col < parameters:
            mcmc_h5 = Path(cache.settings.get('mcmc_h5', None))
            mcmcDir = mcmc_h5.parent
            mle_h5 = mcmcDir / "mle.h5"

            data = cadet.H5()
            data.filename = mle_h5.as_posix()
            data.load()
            multiprocessing.get_logger().info('%s', list(data.root.keys()))
            stat_MLE = data.root.stat_MLE.reshape(1, -1)
            previousResults = numpy.hstack([previousResults, numpy.repeat(stat_MLE, row, 0)])
            multiprocessing.get_logger().info('row: %s  col:%s   shape: %s', row, col, previousResults.shape)

        population = get_population(previousResults, populationSize, diff=0.1)
        checkpoint['starting_population'] = [util.convert_individual_inverse(i, cache) for i in population]
        multiprocessing.get_logger().info('p_burn startup population: %s', population)
        multiprocessing.get_logger().info('p_burn startup: %s', checkpoint['starting_population'])
    else:
        checkpoint['starting_population'] = SALib.sample.sobol_sequence.sample(populationSize, parameters)
    checkpoint['p_burn'] = checkpoint['p_bounds'] = checkpoint['starting_population']

def getCheckPoint(checkpointFile, cache):
    if checkpointFile.exists():
        with checkpointFile.open('rb') as cp_file:
            checkpoint = pickle.load(cp_file)
    else:
        parameters = len(cache.MIN_VALUE)
        
        MCMCpopulationSet = cache.settings.get('MCMCpopulationSet', None)
        if MCMCpopulationSet is not None:
            populationSize = MCMCpopulationSet
        else:
            populationSize = parameters * cache.settings['MCMCpopulation']

        #Population must be even
        populationSize = populationSize + populationSize % 2
        
        #due to emcee 3.0 and RedBlueMove there is a minimum population size to work correctly based on the number of paramters
        populationSize = max(parameters*2, populationSize)

        checkpoint = {}
        checkpoint['state'] = 'start'
        checkpoint['populationSize'] = populationSize
        resetPopulation(checkpoint, cache)
        

        checkpoint['ln_prob_bounds'] = None
        checkpoint['rstate_bounds'] = None
        checkpoint['idx_bounds'] = 0

        checkpoint['ln_prob_burn'] = None
        checkpoint['rstate_burn'] = None
        checkpoint['idx_burn'] = 0
        
        checkpoint['p_chain'] = None
        checkpoint['ln_prob_chain'] = None
        checkpoint['rstate_chain'] = None
        checkpoint['idx_chain'] = 0

        checkpoint['sampler_iterations'] = 0
        checkpoint['sampler_naccepted'] = numpy.zeros(populationSize)

        checkpoint['converge'] = numpy.ones(cache.settings.get('burnStable', 50)) * numpy.nan
    
    checkpoint['length_chain'] = cache.settings.get('chainLength', 50000)
    checkpoint['length_burn'] = cache.settings.get('burnIn', 50000)
    return checkpoint

def setupDEAP(cache, fitness, grad_fitness, grad_search, map_function, creator, base, tools):
    "setup the DEAP variables"
    creator.create("FitnessMax", base.Fitness, weights=[1.0] * cache.numGoals)
    creator.create("Individual", array.array, typecode="d", fitness=creator.FitnessMax, strategy=None, mean=None, confidence=None)

    creator.create("FitnessMaxMeta", base.Fitness, weights=[1.0, 1.0, 1.0, -1.0])
    creator.create("IndividualMeta", array.array, typecode="d", fitness=creator.FitnessMaxMeta, strategy=None)
    cache.toolbox.register("individualMeta", util.initIndividual, creator.IndividualMeta, cache)

    cache.toolbox.register("individual", util.generateIndividual, creator.Individual,
        len(cache.MIN_VALUE), cache.MIN_VALUE, cache.MAX_VALUE, cache)

    cache.toolbox.register("individual_guess", util.initIndividual, creator.Individual, cache)

    cache.toolbox.register('map', map_function)

def process(population_order, cache, halloffame, meta_hof, grad_hof, result_data, results, writer, csv_file):
    if 'gen' not in process.__dict__:
        process.gen = 0

    if 'sim_start' not in process.__dict__:
        process.sim_start = time.time()

    if 'generation_start' not in process.__dict__:
        process.generation_start = time.time()

    population_lookup = {}
    fitnesses_lookup = {}

    log_likelihoods_lookup = {}

    for ll, theta, fit, csv_line, result, individual in results:
        log_likelihoods_lookup[tuple(individual)] = float(ll)
        if result is not None:
            parameters = theta
            fitnesses_lookup[tuple(individual)] = (fit, csv_line, result, tuple(individual))

            ind = cache.toolbox.individual_guess(parameters)
            population_lookup[tuple(individual)] = ind
            
            #result_data['mcmc_score'].append(ll)

    #everything above is async (unordered) and needs to be reordered based on the population_order
    population = [population_lookup[tuple(row)] for row in population_order]
    fitnesses = [fitnesses_lookup[tuple(row)] for row in population_order]
    log_likelihoods = [log_likelihoods_lookup[tuple(row)] for row in population_order]


    stalled, stallWarn, progressWarn = util.process_population(cache.toolbox, cache, population, 
                                                          fitnesses, writer, csv_file, 
                                                          halloffame, meta_hof, process.gen, result_data)
    
    avg, bestMin, bestProd = util.averageFitness(population, cache)

    util.writeProgress(cache, process.gen, population, halloffame, meta_hof, grad_hof, avg, bestMin, bestProd, 
                       process.sim_start, process.generation_start, result_data, line_log=False)

    util.graph_process(cache, process.gen)

    process.gen += 1
    process.generation_start = time.time()
    return log_likelihoods

def process_chain(chain, cache, idx):
    chain_shape = chain.shape
    flat_chain = chain.reshape(chain_shape[0] * chain_shape[1], chain_shape[2])

    flat_chain_transform = util.convert_population(flat_chain, cache)
    chain_transform = flat_chain_transform.reshape(chain_shape)

    return chain, flat_chain, chain_transform, flat_chain_transform

def writeMCMC(cache, bounds_chain, train_chain, chain, bounds_seq, burn_seq, chain_seq, parameters, train_chain_stat, run_chain_stat, tau_percent, iat=None):
    "write out the mcmc data so it can be plotted"
    mcmcDir = Path(cache.settings['resultsDirMCMC'])
    mcmc_h5 = mcmcDir / "mcmc.h5"

    bounds_chain, bounds_chain_flat, bounds_chain_transform, bounds_chain_flat_transform = process_chain(bounds_chain, cache, len(bounds_seq)-1)
    interval_chain = None
    interval_chain_transform = None

    if train_chain is not None:
        train_chain, train_chain_flat, train_chain_transform, train_chain_flat_transform = process_chain(train_chain, cache, len(burn_seq)-1)
        interval_chain = train_chain_flat
        interval_chain_transform = train_chain_flat_transform

    if chain is not None:
        chain, chain_flat, chain_transform, chain_flat_transform = process_chain(chain, cache, len(chain_seq)-1)
        interval_chain = chain_flat
        interval_chain_transform = chain_flat_transform

    if interval_chain is not None:
        flat_interval = interval(interval_chain, cache)
        flat_interval_transform = interval(interval_chain_transform, cache)

        flat_interval.to_csv(mcmcDir / "percentile.csv")
        flat_interval_transform.to_csv(mcmcDir / "percentile_transform.csv")

    h5 = cadet.H5()
    h5.filename = mcmc_h5.as_posix()

    if tau_percent is not None:
        h5.root.tau_percent = tau_percent.reshape(-1, 1)

    if train_chain_stat is not None:
        train_chain_stat, _, train_chain_stat_transform, _ = process_chain(train_chain_stat, cache, len(burn_seq)-1)

        h5.root.train_chain_stat = train_chain_stat
        h5.root.train_chain_stat_transform = train_chain_stat_transform

    if run_chain_stat is not None:
        run_chain_stat, _, run_chain_stat_transform, _ = process_chain(run_chain_stat, cache, len(burn_seq)-1)

        h5.root.run_chain_stat = run_chain_stat
        h5.root.run_chain_stat_transform = run_chain_stat_transform

    if bounds_seq:
        h5.root.bounds_acceptance = numpy.array(bounds_seq).reshape(-1, 1)

    if burn_seq:
        h5.root.burn_in_acceptance = numpy.array(burn_seq).reshape(-1, 1)

    if chain_seq:
        h5.root.mcmc_acceptance = numpy.array(chain_seq).reshape(-1, 1)

    if iat is not None:
        h5.root.integrated_autocorrelation_time = numpy.array(iat)

    if chain is not None:
        h5.root.full_chain = chain
        h5.root.full_chain_transform = chain_transform
        h5.root.flat_chain = chain_flat
        h5.root.flat_chain_transform = chain_flat_transform

    if train_chain is not None:
        h5.root.train_full_chain = train_chain
        h5.root.train_full_chain_transform = train_chain_transform
        h5.root.train_flat_chain = train_chain_flat
        h5.root.train_flat_chain_transform = train_chain_flat_transform

    if bounds_chain is not None:
        h5.root.bounds_full_chain = bounds_chain
        h5.root.bounds_full_chain_transform = bounds_chain_transform
        h5.root.bounds_flat_chain = bounds_chain_flat
        h5.root.bounds_flat_chain_transform = bounds_chain_flat_transform

    if interval_chain_transform is not None:
        mean = numpy.mean(interval_chain_transform,0)
        labels = [5, 10, 50, 90, 95]
        percentile = numpy.percentile(interval_chain_transform, labels, 0)

        h5.root.percentile['mean'] = mean
        for idx, label in enumerate(labels):
            h5.root.percentile['percentile_%s' % label] = percentile[idx,:]
        
    h5.save()

def interval(flat_chain, cache):
    mean = numpy.mean(flat_chain,0)

    percentile = numpy.percentile(flat_chain, [5, 10, 50, 90, 95], 0)

    data = numpy.vstack( (mean, percentile) ).transpose()

    pd = pandas.DataFrame(data, columns = ['mean', '5', '10', '50', '90', '95'])
    pd.insert(0, 'name', cache.parameter_headers_actual)
    pd.set_index('name')
    return pd


import sys

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

from cadet import H5

from CADETMatch.cache import cache

from pathlib import Path
import numpy

def main():
    cache.setup(sys.argv[1])
    
    mcmcDir = Path(cache.settings['resultsDirMCMC'])

    h5_data = H5()
    h5_data.filename = (mcmcDir / "kde_settings.h5").as_posix()
    h5_data.load()

    store = h5_data.root.store

    mcmcDir = Path(cache.settings['resultsDirMCMC'])
    plt.figure(figsize=[10,10])
    plt.scatter(numpy.log10(store[:,0]), numpy.log10(store[:,1]))
    plt.xlabel('bandwidth')
    plt.ylabel('cross_val_score')
    plt.savefig(str(mcmcDir / "log_bandwidth.png" ), bbox_inches='tight')
    plt.close()

    plt.figure(figsize=[10,10])
    plt.scatter(store[:,0], store[:,1])
    plt.xlabel('bandwidth')
    plt.ylabel('cross_val_score')
    plt.savefig(str(mcmcDir / "bandwidth.png" ), bbox_inches='tight')
    plt.close()

    mcmcDir = Path(cache.settings['resultsDirMCMC'])

    h5_data = H5()
    h5_data.filename = (mcmcDir / "kde_settings.h5").as_posix()
    h5_data.load()

    dir_base = cache.settings.get('resultsDirBase')
    file = dir_base / 'kde_data.h5'

    kde_data = H5()
    kde_data.filename = file.as_posix()
    kde_data.load()

    times = {}
    values = {}
    for key in kde_data.root.keys():
        if '_time' in key:
            times[key.replace('_time', '')] = kde_data.root[key]
        if '_unit' in key:
            values[key] = kde_data.root[key]

    error_model = mcmcDir / "error_model"
    error_model.mkdir(parents=True, exist_ok=True)

    for key,value in values.items():
        time = times[key.split('_unit', 1)[0]]
        plt.figure(figsize=[10,10])
        row, col = value.shape
        alpha = max(1.0/row, 0.01)
        plt.plot(time, value.T, 'g', alpha=alpha)
        plt.xlabel('time')
        plt.ylabel('concentration')
        plt.savefig((error_model / ("%s.png" % key) ).as_posix(), bbox_inches='tight')
        plt.close()

if __name__ == "__main__":
    main()





import CADETMatch.util as util

name = "sum"
count = 0
count_extended = 0

def getUnit(location):
    return location.split('/')[3]

def transform(parameter):
    return []

def untransform(seq, cache, parameter):
    return [], []

def untransform_matrix(matrix, cache, parameter):
    return None

untransform_matrix_inputorder = untransform_matrix

def getValue(sim, location, bound=None, comp=None, index=None):
    if bound is not None:
        unit = getUnit(location)
        boundOffset = util.getBoundOffset(sim.root.input.model[unit])

        if comp == -1:
            position = ()
            return sim[location.lower()]
        else:
            position = boundOffset[comp] + bound
            return sim[location.lower()][position]

    if index is not None:
        if index == -1:
            return sim[location.lower()]
        else:
            return sim[location.lower()][index]

def setValue(sim, value, location, bound=None, comp=None, index=None):
    if bound is not None:
        unit = getUnit(location)
        boundOffset = util.getBoundOffset(sim.root.input.model[unit])

        if comp == -1:
            position = ()
            sim[location.lower()] = value
        else:
            position = boundOffset[comp] + bound
            sim[location.lower()][position] = value

    if index is not None:
        if index == -1:
            sim[location.lower()] = value
        else:
            sim[location.lower()][index] = value
        
def setSimulation(sim, parameter, seq, cache, experiment):
    if parameter.get('experiments', None) is None or experiment['name'] in parameter['experiments']:
        location1 = parameter['location1']
        location2 = parameter['location2']
        locationSum = parameter['locationSum']
    
        try:
            comp1 = parameter['component1']
            bound1 = parameter['bound1']
            index1 = None
        except KeyError:
            index1 = parameter['index1']
            bound1 = None
            comp1 = None
        value1 = getValue(sim, location1, bound=bound1, comp=comp1, index=index1)

        try:
            comp2 = parameter['component2']
            bound2 = parameter['bound2']
            index2 = None
        except KeyError:
            index2 = parameter['index2']
            bound2 = None
            comp2 = None

        value2 = getValue(sim, location2, bound=bound2, comp=comp2, index=index2)

        try:
            compSum = parameter['componentSum']
            boundSum = parameter['boundSum']
            indexSum = None
        except KeyError:
            indexSum = parameter['indexSum']
            boundSum = None
            compSum = None
        setValue(sim, value1+value2, locationSum, bound=boundSum, comp=compSum, index=indexSum)

    return [],[]

def setupTarget(parameter):
    return [], 0

def getBounds(parameter):
    return None,None

def getHeaders(parameter):
    return []

def getHeadersActual(parameter):
    return []

def setBounds(parameter, lb, ub):
    return None






"Cache module. Used to store per worker information so it does not need to be recalculated"

from pathlib import Path
import json
from cadet import Cadet

import numpy

import CADETMatch.plugins as plugins
import os
from deap import tools
from sklearn import preprocessing

import multiprocessing

class Cache:
    def __init__(self):
        self.settings = None
        self.headers = None
        self.numGoals = None
        self.target = None
        self.MIN_VALUE = None
        self.MAX_VALUE = None
        self.badScore = None
        self.WORST = None
        self.json_path = None
        self.adaptive = None
        self.parameter_indexes = None
        self.score_indexes = None
        self.score_headers = None
        self.graphGenerateTime = None
        self.graphMetaTime = None
        self.lastGraphTime = None
        self.lastMetaTime = None
        self.metaResultsOnly = 1
        self.stallGenerations = 10
        self.stallCorrect = 5
        self.progressCorrect = 5
        self.lastProgressGeneration = -1
        self.generationsOfProgress = 0
        self.fullTrainingData = 0
        self.normalizeOutput = True
        self.sobolGeneration = True
        self.graphSpearman = False
        self.continueMCMC = False
        self.errorBias = True
        self.altFeatures = False
        self.altFeatureNames = []
        self.progress_headers = ['Generation', 'Population', 'Dimension In', 'Dimension Out', 'Search Method',
                                 'Pareto Front', 'Product Score', 'Minimum Score', 'Average Score',
                                 'Pareto Mean Product Score', 'Pareto Mean Minimum Score', 'Pareto Mean Average Score',
                                 'Pareto Meta Product Score', 'Pareto Meta Min Score', 'Pareto Meta Mean Score',
                                 'Elapsed Time', 'Generation Time', 'Total CPU Time', 'Last Progress Generation',
                                 'Generations of Progress']

    def setup_dir(self, json_path):
        self.json_path = json_path
        self.setupSettings()

        baseDir = self.settings.get('baseDir', None)
        if baseDir is not None:
            os.chdir(baseDir)
            self.settings['resultsDir'] = Path(baseDir) / self.settings['resultsDir']

        #create used paths in settings, only the root process will make the directories later
        self.settings['resultsDirEvo'] = Path(self.settings['resultsDir']) / "evo"
        self.settings['resultsDirMeta'] = Path(self.settings['resultsDir']) / "meta"
        self.settings['resultsDirGrad'] = Path(self.settings['resultsDir']) / "grad"
        self.settings['resultsDirMisc'] = Path(self.settings['resultsDir']) / "misc"
        self.settings['resultsDirSpace'] = Path(self.settings['resultsDir']) / "space"
        self.settings['resultsDirProgress'] = Path(self.settings['resultsDir']) / "progress"
        self.settings['resultsDirLog'] = Path(self.settings['resultsDir']) / "log"
        self.settings['resultsDirMCMC'] = Path(self.settings['resultsDir']) / "mcmc"
        self.settings['resultsDirBase'] = Path(self.settings['resultsDir'])

    def setup(self, json_path, load_plugins=True):
        "setup the cache based on the json file being used"
        if load_plugins:
            self.scores = plugins.get_plugins('scores')
            self.search = plugins.get_plugins('search')
            self.transforms = plugins.get_plugins('transform')

        if json_path != self.json_path:
            self.json_path = json_path
            self.setupSettings()

            baseDir = self.settings.get('baseDir', None)
            if baseDir is not None:
                os.chdir(baseDir)
                self.settings['resultsDir'] = Path(baseDir) / self.settings['resultsDir']

            self.settings['resultsDirEvo'] = Path(self.settings['resultsDir']) / "evo"
            self.settings['resultsDirMeta'] = Path(self.settings['resultsDir']) / "meta"
            self.settings['resultsDirGrad'] = Path(self.settings['resultsDir']) / "grad"
            self.settings['resultsDirMisc'] = Path(self.settings['resultsDir']) / "misc"
            self.settings['resultsDirSpace'] = Path(self.settings['resultsDir']) / "space"
            self.settings['resultsDirProgress'] = Path(self.settings['resultsDir']) / "progress"
            self.settings['resultsDirLog'] = Path(self.settings['resultsDir']) / "log"
            self.settings['resultsDirMCMC'] = Path(self.settings['resultsDir']) / "mcmc"
            self.settings['resultsDirBase'] = Path(self.settings['resultsDir'])

        self.abstolFactor = self.settings.get('abstolFactor', 1e-4)
        self.reltol = self.settings.get('reltol', 1e-4)
        self.abstolFactorGrad = self.settings.get('abstolFactorGrad', 1e-8)
        self.reltolGrad = self.settings.get('reltolGrad', 1e-8)
        self.dynamicTolerance = bool(self.settings.get('dynamicTolerance', False))

        self.errorBias = bool(self.settings.get('errorBias', True))

        Cadet.cadet_path = self.settings['CADETPath']

        self.normalizeOutput = bool(self.settings.get('normalizeOutput', False))
        self.connectionNumberEntries = int(self.settings.get('connectionNumberEntries', 5))

        self.setupHeaders()
        self.setupTarget()
        self.setupMinMax()

        self.WORST = [self.badScore] * self.numGoals

        self.settings['transform'] = self.transform

        self.correct = None
        if "correct" in self.settings:
            self.correct = numpy.array([f(v) for f, v in zip(self.settings['transform'], self.settings['correct'])])
            
        self.error_path = Path(self.settings['resultsDirBase'], "error.csv")

        self.graphGenerateTime = int(self.settings.get('graphGenerateTime', 3600))
        self.graphMetaTime = int(self.settings.get('graphMetaTime', 1200))

        self.metaResultsOnly = self.settings.get('metaResultsOnly', 1)
        self.stallGenerations = int(self.settings.get('stallGenerations', 10))
        self.stallCorrect = int(self.settings.get('stallCorrect', 5))
        self.progressCorrect = int(self.settings.get('progressCorrect', 5))

        self.fullTrainingData = int(self.settings.get('fullTrainingData', 0))

        self.sobolGeneration = bool(self.settings.get('soboloGeneration', True)) or bool(self.settings.get('sobolGeneration', True))
        self.graphSpearman = bool(self.settings.get('graphSpearman', False))
        self.scoreMCMC = self.settings.get('scoreMCMC', "sse")

        self.continueMCMC = bool(self.settings.get('continueMCMC', False))
        self.MCMCTauMult = int(self.settings.get('MCMCTauMult', 50))

        self.cross_eta = int(self.settings.get('cross_eta', 30))
        self.mutate_eta = int(self.settings.get('mutate_eta', 70))

        self.gradVector = bool(self.settings.get('gradVector', 0))

        self.tempDir = self.settings.get('tempDir', None)
        self.graphType = self.settings.get('graphType', 1)

        self.checkpointInterval = self.settings.get('checkpointInterval', 600)
        self.setupMetaMask()

        self.debugWrite = bool(self.settings.get('debugWrite', False))

        self.finalGradRefinement = bool(self.settings.get('finalGradRefinement', False))

        self.multiStartPercent = self.settings.get('multiStartPercent', 0.1)

        if "MCMCpopulation" not in self.settings:
            self.settings['MCMCpopulation'] = self.settings['population']


        if self.numGoals == 1:
            #with one goal one of the emo functions breaks, this is a temporary fix

            def sortNDHelperB(best, worst, obj, front):
                if obj < 0:
                    return
                sortNDHelperB(best, worst, obj, front)

            tools.emo.sortNDHelperB = sortNDHelperB

    def resetTransform(self, json_path):
        if json_path != self.json_path:
            self.json_path = json_path

            settings_file = Path(self.json_path)
            with settings_file.open() as json_data:
                settings = json.load(json_data)

                self.settings['parameters'] = settings['parameters']

                self.setupHeaders()
                self.setupTarget()
                self.setupMinMax()

                self.WORST = [self.badScore] * self.numGoals

                self.settings['transform'] = self.transform

                self.correct = None
                if "correct" in self.settings:
                    self.correct = numpy.array([f(v) for f, v in zip(self.settings['transform'], self.settings['correct'])])


    def setupMetaMask(self):
        meta_mask_seq = []

        for idx, experiment in enumerate(self.settings['experiments']):
            for feature in experiment['features']:
                if feature['type'] in self.scores:
                    settings = self.scores[feature['type']].get_settings(feature)
                    meta_mask = settings.meta_mask
                    count = settings.count
                    #multiprocessing.get_logger().info('%s %s %s %s', idx, feature, meta_mask, count)

                    meta_mask_seq.extend([meta_mask,] * count)
        #multiprocessing.get_logger().info("%s", meta_mask_seq)
        self.meta_mask = numpy.array(meta_mask_seq)
        
    def setupSettings(self):
        settings_file = Path(self.json_path)
        with settings_file.open() as json_data:
            self.settings = json.load(json_data)

            if 'CSV' in self.settings:
                self.settings['csv'] = self.settings['CSV']
            if 'csv' not in self.settings:
                self.settings['csv'] = 'results.csv'

            if self.settings['searchMethod'] == 'Gradient':
                self.settings['population'] = 1
            else:
                self.settings['population'] = int(self.settings['population'])
            
            self.settings['maxPopulation'] = int(self.settings.get('maxPopulation', self.settings['population']))
            self.settings['minPopulation'] = int(self.settings.get('minPopulation', self.settings['population']))


            if "bootstrap" in self.settings:
                self.settings['bootstrap']['samples'] = int(self.settings['bootstrap']['samples'])

    def setupHeaders(self):
        self.headers = ['Time', 'Name', 'Method', 'Condition Number',]

        self.numGoals = 0
        self.badScore = 0.0

        base = len(self.headers)

        parameter_headers = []
        parameter_headers_actual = []
        
        for parameter in self.settings['parameters']:
            parameter_headers.extend(self.transforms[parameter['transform']].getHeaders(parameter))
            parameter_headers_actual.extend(self.transforms[parameter['transform']].getHeadersActual(parameter))

        self.parameter_headers = parameter_headers
        self.parameter_headers_actual = parameter_headers_actual

        self.headers.extend(parameter_headers)

        parameters = len(self.headers)
        self.parameter_indexes = list(range(base, parameters))

        self.score_headers = []

        badScore = []

        for idx, experiment in enumerate(self.settings['experiments']):
            experimentName = experiment['name']
            experiment['headers'] = []
            for feature in experiment['features']:
                if feature['type'] in self.scores:
                    temp = self.scores[feature['type']].headers(experimentName, feature)

                    settings = self.scores[feature['type']].get_settings(feature)

                    if settings.meta_mask:
                        self.numGoals += len(temp)
                    badScore.append(settings.badScore)

                    self.score_headers.extend(temp)
                    experiment['headers'].extend(temp)

        self.badScore = min(badScore)

        self.headers.extend(self.score_headers)                      
        
        self.meta_headers = ['Product Root Score', 'Min Score', 'Mean Score', 'SSE']

        self.headers.extend(self.meta_headers)

        scores = len(self.headers)
        self.score_indexes = list(range(parameters, scores))

    def add_units_isotherm(self, units_used, isotherm):
        if not isinstance(isotherm, list):
           isotherm = [isotherm,]
        for path in isotherm:
            for element in path.split('/'):
                if element.startswith('unit_'):
                    if element not in units_used:
                        units_used.append(element)

    def add_units_features_alt(self, units_used, features_alt):
        for feature_alt in features_alt:
            self.add_units_isotherm(units_used, feature_alt.get('isotherm', ''))
            for feature in feature_alt['features']:
                self.add_units_isotherm(units_used, feature.get('isotherm', ''))
                self.add_units_isotherm(units_used, feature.get('unit_name', ''))

    def add_units_error_model(self, kde_synthetic, target):
        for error_model in kde_synthetic:
            error_model_name = error_model['name']
            if error_model_name in target:
                units_used = target[error_model_name]['units_used']
                for unit in error_model['units']:
                    unit_name = 'unit_%03d' % int(unit)
                    if unit_name not in units_used:
                        units_used.append(unit_name)

    def setupTarget(self):
        self.target = {}
        self.adaptive = True

        for experiment in self.settings['experiments']:
            self.target[experiment["name"]] = self.setupExperiment(experiment)
        if 'kde_synthetic' in self.settings:
            self.add_units_error_model(self.settings['kde_synthetic'], self.target)
        self.target['bestHumanScores'] = numpy.ones(5) * self.badScore

        #SSE are negative so they sort correctly with better scores being less negative
        self.target['bestHumanScores'][4] = self.badScore;  

        #setup sensitivities
        parms = []
        sensitivityOk = 1
        for parameter in self.settings['parameters']:
            try:
                comp = parameter['component']
            except KeyError:
                sensitivityOk = 0
                break

            transform = parameter['transform']

            sens_parms, sensitivityOk = self.transforms[transform].setupTarget(parameter)
            parms.extend(sens_parms)

        if sensitivityOk:
            self.target['sensitivities'] = parms
        else:
            self.target['sensitivities'] = []

    def setupExperiment(self, experiment, sim=None, dataFromSim=0):
        temp = {}

        units_used = []
        
        if sim is None:
            sim = Cadet()
            sim.filename = Path(experiment['HDF5']).as_posix()
            sim.load()
            sim.root.experiment_name = experiment['name']

        abstol = sim.root.input.solver.time_integrator.abstol

        conn = sim.root.input.model.connections.switch_000.connections

        conn = numpy.array(conn)
        conn = numpy.reshape(conn, [-1, self.connectionNumberEntries])

        #find all the entries that connect to the column
        filter = conn[:, 1] == 1

        #flow is the sum of all flow rates that connect to this column which is in the last column
        flow = sum(conn[filter, -1])

        if sim.root.input.model.unit_001.unit_type == b'CSTR':
            volume = float(sim.root.input.model.unit_001.init_volume)

            CV_time = volume / flow

        else:
            #CV needs to be based on superficial velocity not interstitial velocity
            length = float(sim.root.input.model.unit_001.col_length)

            velocity = sim.root.input.model.unit_001.velocity
            if velocity == {}:
                velocity = 1.0
            velocity = float(velocity)

            area = sim.root.input.model.unit_001.cross_section_area
            if area == {}:
                area = 1.0
            area = float(area)

            if area == 1 and abs(velocity) != 1:
                CV_time = length / velocity
            else:
                CV_time = (area * length) / flow

        #force CSV to lowercase
        if 'CSV' in experiment:
            experiment['csv'] = experiment['CSV']

        if dataFromSim:
            temp['time'], temp['value'] = get_times_values(sim, {'isotherm':experiment['isotherm']})

        elif 'csv' in experiment:
            data = numpy.loadtxt(experiment['csv'], delimiter=',')

            temp['time'] = data[:, 0]
            temp['value'] = data[:, 1]

            if self.normalizeOutput:
                temp['factor'] = 1.0/numpy.max(temp['value'])
            else:
                temp['factor'] = 1.0
            temp['valueFactor'] = temp['value'] * temp['factor']

        if "featuresAlt" in experiment:
            self.altFeatures = True
            self.altFeatureNames = [altFeature['name'] for altFeature in experiment['featuresAlt']]

            self.add_units_features_alt(units_used, experiment['featuresAlt'])

        peak_maxes = []
        for feature in experiment['features']:
            featureName = feature['name']
            featureType = feature['type']
            
            temp[featureName] = feature

            #switch to lower case
            if 'CSV' in feature:
                feature['csv'] = feature['CSV']

            if 'csv' in feature:
                if dataFromSim:
                    temp[featureName]['time'], temp[featureName]['value'] = get_times_values(sim, {'isotherm':feature['isotherm']})
                else:
                    dataLocal = numpy.loadtxt(feature['csv'], delimiter=',')
                    temp[featureName]['time'] = dataLocal[:, 0]
                    temp[featureName]['value'] = dataLocal[:, 1]
            else:
                temp[featureName]['time'] = temp['time']
                temp[featureName]['value'] = temp['value']

            if self.normalizeOutput:
                temp[featureName]['factor'] = 1.0/numpy.max(temp[featureName]['value'])
                temp[featureName]['value'] = temp[featureName]['value'] * temp[featureName]['factor']
            else:
                temp[featureName]['factor'] = 1.0

            try:
                featureStart = float(feature['start'])
                featureStop = float(feature['stop'])
            except KeyError:
                feature['start']= featureStart = temp[featureName]['time'][0]
                feature['stop'] = featureStop = temp[featureName]['time'][-1]

            if 'isotherm' in feature:
                temp[featureName]['isotherm'] = feature['isotherm']
            else:
                temp[featureName]['isotherm'] = experiment['isotherm']

            self.add_units_isotherm(units_used, temp[featureName]['isotherm'])

            temp[featureName]['selected'] = (temp[featureName]['time'] >= featureStart) & (temp[featureName]['time'] <= featureStop)
            
            selectedTimes = temp[featureName]['time'][temp[featureName]['selected']]
            selectedValues = temp[featureName]['value'][temp[featureName]['selected']]

            if "unit_name" in feature:
                self.add_units_isotherm(units_used, feature['unit_name'])

            if featureType in self.scores:
                temp[featureName].update(self.scores[featureType].setup(sim, feature, selectedTimes, selectedValues, CV_time, abstol, self))
                settings = self.scores[featureType].get_settings(feature)
                self.adaptive = settings.adaptive
                if 'peak_max' in temp[featureName]:
                    peak_maxes.append(temp[featureName]['peak_max']/temp[featureName]['factor'])
         
        temp['smallest_peak'] = min(peak_maxes)   
        temp['units_used'] = units_used
        return temp

    def setupMinMax(self):
        "build the minimum and maximum parameter boundaries"
        self.MIN_VALUE = []
        self.MAX_VALUE = []
        #self.MIN_VALUE_GRAD = []
        #self.MAX_VALUE_GRAD = []
        self.transform = []
        #self.grad_transform = []

        for parameter in self.settings['parameters']:
            transform = parameter['transform']
            minValues, maxValues = self.transforms[transform].getBounds(parameter)
            #minGradValues, maxGradValues = self.transforms[transform].getGradBounds(parameter)
            transforms = self.transforms[transform].transform(parameter)
            #grad_transforms = self.transforms[transform].grad_transform(parameter)

            if minValues:
                self.MIN_VALUE.extend(minValues)
                self.MAX_VALUE.extend(maxValues)
                self.transform.extend(transforms)
                #self.grad_transform.extend(grad_transforms)
                #self.MIN_VALUE_GRAD.extend(minGradValues)
                #self.MAX_VALUE_GRAD.extend(maxGradValues)

cache = Cache()


def get_times_values(simulation, target):
    "simplified version of the function so that util does not have to be imported"
    times = simulation.root.output.solution.solution_times
    isotherm = target['isotherm']

    if isinstance(isotherm, list):
        values = numpy.sum([simulation[i] for i in isotherm], 0)
    else:
        values = simulation[isotherm]
        
    return times, values



import scipy.stats
import numpy
import scipy.optimize
import scipy.interpolate
import scipy.signal
import numpy.linalg
import CADETMatch.calc_coeff as calc_coeff
import multiprocessing
import sys
import SALib.sample.sobol_sequence

def roll(x, shift):
    if shift > 0:
        temp = numpy.pad(x,(shift,0), mode='constant')
        return temp[:-shift]
    elif shift < 0:
        temp = numpy.pad(x,(0,numpy.abs(shift)), mode='constant')
        return temp[numpy.abs(shift):]
    else:
        return x


def roll_spline(times, values, shift):
    "this function does approximately what the roll function does except that is used the spline so the shift does not have to be an integer and the points are resampled"

    spline = scipy.interpolate.InterpolatedUnivariateSpline(times, values, ext=3)

    times_new = times - shift

    values_new = spline(times_new)

    return values_new


def cross_correlate(exp_time_values, sim_data_values, exp_data_values):
    corr = scipy.signal.correlate(exp_data_values, sim_data_values)/(numpy.linalg.norm(sim_data_values) * numpy.linalg.norm(exp_data_values))

    #need +1 due to how correlate works
    index = numpy.argmax(corr) + 1

    roll_index = index - len(exp_time_values)

    score = corr[index]

    sim_time_values = roll(exp_time_values, shift=int(numpy.ceil(roll_index)))

    diff_time = numpy.abs(exp_time_values[int(len(exp_time_values)/2)] - sim_time_values[int(len(exp_time_values)/2)])

    return score, diff_time

def pearson(exp_time_values, sim_data_values, exp_data_values):
    corr = scipy.signal.correlate(exp_data_values, sim_data_values)/(numpy.linalg.norm(sim_data_values) * numpy.linalg.norm(exp_data_values))

    #need +1 due to how correlate works
    index = numpy.argmax(corr) + 1

    roll_index = index - len(exp_time_values)

    try:
        score = corr[index]
    except IndexError:
        score = 0.0
        multiprocessing.get_logger().warn("Index error in pearson score at index %s out of %s entries", index, len(corr))

    endTime = exp_time_values[-1]

    sim_time_values = roll(exp_time_values, shift=int(numpy.ceil(roll_index)))

    diff_time = numpy.abs(exp_time_values[int(len(exp_time_values)/2)] - sim_time_values[int(len(exp_time_values)/2)])

    sim_data_values_copy = roll(sim_data_values, shift=int(numpy.ceil(roll_index)))

    try:
        pear = scipy.stats.pearsonr(exp_data_values, sim_data_values_copy)[0]
    except ValueError:
        multiprocessing.get_logger().warn("Pearson correlation failed to do NaN or InF in array  exp_array: [%s]   sim_array: [%s]", 
                                          list(exp_data_values), list(sim_data_values_copy))
        pear = 0
    score = pear_corr(pear)
    
    return score, diff_time

def refine_time(dt, exp_time_values, sim_data_values, exp_data_values):
    "refine the time using powells method (gradient free) system is not smooth enough for gradient to work well"
    spline = scipy.interpolate.InterpolatedUnivariateSpline(exp_time_values, sim_data_values, ext=3)

    exp_time_values = numpy.array(exp_time_values)
    exp_data_values = numpy.array(exp_data_values)
    
    def goal_sse(offset):
        sim_data_values_copy = spline(exp_time_values - offset)

        sse = numpy.sum((exp_data_values - sim_data_values_copy)**2)

        return sse
    
    result = scipy.optimize.minimize(goal_sse, dt, method='powell')
    
    return result.x

def pearson_spline(exp_time_values, sim_data_values, exp_data_values):
    #resample to a much smaller time step to get a more precise offset
    dt = 1e-1
    times = numpy.arange(exp_time_values[0], exp_time_values[-1], dt)
    
    sim_resample = scipy.interpolate.InterpolatedUnivariateSpline(exp_time_values, sim_data_values, ext=3)(times)
    exp_resample = scipy.interpolate.InterpolatedUnivariateSpline(exp_time_values, exp_data_values, ext=3)(times)    
    
    corr = scipy.signal.correlate(exp_resample, sim_resample)  #/(numpy.linalg.norm(sim_resample) * numpy.linalg.norm(exp_resample))

    index = numpy.argmax(corr)
    
    dt_pre = (index - len(times) + 1) * dt

    #refine the time offset using a deterministic method (powells)
    dt = refine_time(dt_pre, exp_time_values, sim_data_values, exp_data_values)
    
    #calculate pearson correlation at the new time
    spline = scipy.interpolate.InterpolatedUnivariateSpline(exp_time_values, sim_data_values, ext=3)
    sim_data_values_copy = spline(exp_time_values - dt)
    try:
        pear = scipy.stats.pearsonr(exp_data_values, sim_data_values_copy)[0]
    except ValueError:
        multiprocessing.get_logger().warn("Pearson correlation failed to do NaN or InF in array  exp_array: [%s]   sim_array: [%s]", 
                                          list(exp_data_values), list(sim_data_values_copy))
        pear = 0
    score = pear_corr(pear)

    return score, dt

def pearson_spline_fun(exp_time_values, exp_data_values, sim_spline):
    #resample to a much smaller time step to get a more precise offset
    dt = 1e-1
    times = numpy.arange(exp_time_values[0], exp_time_values[-1], dt)
    
    sim_resample = sim_spline(times)
    exp_resample = scipy.interpolate.InterpolatedUnivariateSpline(exp_time_values, exp_data_values, ext=3)(times)    
    
    corr = scipy.signal.correlate(exp_resample, sim_resample)  #/(numpy.linalg.norm(sim_resample) * numpy.linalg.norm(exp_resample))

    index = numpy.argmax(corr)
    
    dt_pre = (index - len(times) + 1) * dt

    #refine the time offset using a deterministic method (powells)
    dt = refine_time_fun(dt_pre, exp_time_values, exp_data_values, sim_spline)
    
    #calculate pearson correlation at the new time
    sim_data_values_copy = sim_spline(exp_time_values - dt)
    try:
        pear = scipy.stats.pearsonr(exp_data_values, sim_data_values_copy)[0]
    except ValueError:
        multiprocessing.get_logger().warn("Pearson correlation failed to do NaN or InF in array  exp_array: [%s]   sim_array: [%s]", 
                                          list(exp_data_values), list(sim_data_values_copy))
        pear = 0
    score = pear_corr(pear)
    return score, dt

def refine_time_fun(dt, exp_time_values, exp_data_values, spline):
    "refine the time using powells method (gradient free) system is not smooth enough for gradient to work well"
    exp_time_values = numpy.array(exp_time_values)
    exp_data_values = numpy.array(exp_data_values)
    
    def goal_sse(offset):
        sim_data_values_copy = spline(exp_time_values - offset)

        sse = numpy.sum((exp_data_values - sim_data_values_copy)**2)

        return sse
    
    result = scipy.optimize.minimize(goal_sse, dt, method='powell')
    
    return result.x

def time_function_decay(CV_time, peak_time, diff_input=False):
    x_exp = numpy.array([0, 1.0*CV_time])
    y_exp = numpy.array([1, 0.5])

    #a, b = calc_coeff.exponential_coeff(x_exp[0], y_exp[0], x_exp[1], y_exp[1])

    a, b = calc_coeff.linear_coeff(x_exp[0], y_exp[0], x_exp[1], y_exp[1])
    
    def wrapper(x):

        if diff_input:
            diff = x
        else:
            diff = numpy.abs(x - peak_time)

        #value = max(0.0, calc_coeff.exponential(diff, a, b))
        value = max(0.0, calc_coeff.linear(diff, a, b))

        return value

    return wrapper

def time_function_decay_cv(CV_time, times, peak_time):

    time = numpy.max(numpy.abs(numpy.array([times[-1] - peak_time, peak_time - times[0]])))
    cv_time = time /CV_time
    
    x_exp = numpy.array([0, cv_time])
    y_exp = numpy.array([1, 0.0])

    a, b = calc_coeff.linear_coeff(x_exp[0], y_exp[0], x_exp[1], y_exp[1])
    
    def wrapper(x):
        diff = x/CV_time
        value = max(0.0, calc_coeff.linear(diff, a, b))

        return value

    return wrapper

def time_function_cv(CV_time, times, peak_time):

    time = numpy.max(numpy.abs(numpy.array([times[-1] - peak_time, peak_time - times[0]])))
    cv_time = time /CV_time

    x_lin_1 = numpy.array([0.0, 0.1])
    y_lin_1 = numpy.array([1.0, 0.95])
    
    x_lin_2 = numpy.array([0.1, cv_time])
    y_lin_2 = numpy.array([0.95, 0.0])

    a1, b1 = calc_coeff.linear_coeff(x_lin_1[0], y_lin_1[0], x_lin_1[1], y_lin_1[1])
    a2, b2 = calc_coeff.linear_coeff(x_lin_2[0], y_lin_2[0], x_lin_2[1], y_lin_2[1])
    
    def wrapper(x):
        diff_cv = x/CV_time
        if diff_cv <= 0.1:
            value = max(0.0, calc_coeff.linear(diff_cv, a1, b1))
        else:
            value = max(0.0, calc_coeff.linear(diff_cv, a2, b2))
        return value
    return wrapper


def time_function_decay_exp(CV_time, peak_time, diff_input=False):
    x_exp = numpy.array([0, 1.0*CV_time])
    y_exp = numpy.array([1, 0.5])

    a, b = calc_coeff.exponential_coeff(x_exp[0], y_exp[0], x_exp[1], y_exp[1])

    #a, b = calc_coeff.linear_coeff(x_exp[0], y_exp[0], x_exp[1], y_exp[1])
    
    def wrapper(x):

        if diff_input:
            diff = x
        else:
            diff = numpy.abs(x - peak_time)

        value = max(0.0, calc_coeff.exponential(diff, a, b))
        #value = max(0.0, calc_coeff.linear(diff, a, b))

        return value

    return wrapper

def time_function(CV_time, peak_time, diff_input=False):
    x_lin = numpy.array([0, 4*CV_time])
    #y_lin = numpy.array([1, 0.0])

    y_lin = numpy.array([1, 0.05])

    a, b = calc_coeff.exponential_coeff(x_lin[0], y_lin[0], x_lin[1], y_lin[1])
    #a, b = calc_coeff.linear_coeff(x_lin[0], y_lin[0], x_lin[1], y_lin[1])

    def wrapper(x):

        if diff_input:
            diff = x
        else:
            diff = numpy.abs(x - peak_time)

        #if diff < CV_time/2.0:
        #value = max(0.0, calc_coeff.linear(diff, a, b))
        value = max(0.0, calc_coeff.exponential(diff, a, b))

        return value

    return wrapper

def time_function2(CV_time, peak_time, diff_input=False):
    #x_exp = numpy.array([CV_time/2.0, 10.0*CV_time])
    x_exp = numpy.array([60.0, 2.0*CV_time])
    y_exp = numpy.array([0.97, 0.5])

    #x_lin = numpy.array([0, CV_time/2.0])
    x_lin = numpy.array([0, 60.0])
    y_lin = numpy.array([1, 0.97])

    a_exp, b_exp = calc_coeff.exponential_coeff(x_exp[0], y_exp[0], x_exp[1], y_exp[1])
    a_lin, b_lin = calc_coeff.linear_coeff(x_lin[0], y_lin[0], x_lin[1], y_lin[1])

    def wrapper(x):

        if diff_input:
            diff = x
        else:
            diff = numpy.abs(x - peak_time)

        #if diff < CV_time/2.0:
        if diff < 60.0:
            value = calc_coeff.linear(diff, a_lin, b_lin)
        else:
            value = calc_coeff.exponential(diff, a_exp, b_exp)

        return value

    return wrapper

def value_function(peak_height, tolerance=1e-8, bottom_score = 0.0):
    #if the peak height is 0 or less than the tolerance it needs to be treated as a special case to prevent divide by zero problems
    x = numpy.array([0.0, 1.0])
    y = numpy.array([1.0, bottom_score])

    #a, b = calc_coeff.exponential_coeff(x[0], y[0], x[1], y[1])
    a, b = calc_coeff.linear_coeff(x[0], y[0], x[1], y[1])
    
    if numpy.abs(peak_height) < tolerance:
        multiprocessing.get_logger().warn("peak height less than tolerance %s %s", tolerance, peak_height)
        def wrapper(x):
            if numpy.abs(x) < tolerance:
                return 1.0
            else:
                diff = numpy.abs(x-tolerance)/numpy.abs(tolerance)
                #return max(0, calc_coeff.exponential(diff, a, b))
                return max(0, calc_coeff.linear(diff, a, b))
    else:
        def wrapper(x):
            diff = numpy.abs(x-peak_height)/numpy.abs(peak_height)
            #return max(0, calc_coeff.exponential(diff, a, b))
            return max(0, calc_coeff.linear(diff, a, b))

    return wrapper

def value_function_exp(peak_height, tolerance=1e-8, bottom_score = 0.05):
    #if the peak height is 0 or less than the tolerance it needs to be treated as a special case to prevent divide by zero problems
    x = numpy.array([0.0, 1.0])
    y = numpy.array([1.0, bottom_score])

    a, b = calc_coeff.exponential_coeff(x[0], y[0], x[1], y[1])
    #a, b = calc_coeff.linear_coeff(x[0], y[0], x[1], y[1])
    
    if numpy.abs(peak_height) < tolerance:
        multiprocessing.get_logger().warn("peak height less than tolerance %s %s", tolerance, peak_height)
        def wrapper(x):
            if numpy.abs(x) < tolerance:
                return 1.0
            else:
                diff = numpy.abs(x-tolerance)/numpy.abs(tolerance)
                return max(0, calc_coeff.exponential(diff, a, b))
                #return max(0, calc_coeff.linear(diff, a, b))
    else:
        def wrapper(x):
            diff = numpy.abs(x-peak_height)/numpy.abs(peak_height)
            return max(0, calc_coeff.exponential(diff, a, b))
            #return max(0, calc_coeff.linear(diff, a, b))

    return wrapper

def slope_function(peak_slope):
    #if the peak height is 0 or less than the tolerance it needs to be treated as a special case to prevent divide by zero problems
    x = numpy.array([0.0, 4.0])
    y = numpy.array([1.0, 0.0])

    a, b = calc_coeff.linear_coeff(x[0], y[0], x[1], y[1])
    
    def wrapper(x):
        diff = numpy.abs(x-peak_slope)/numpy.abs(peak_slope)
        return max(0, calc_coeff.linear(diff, a, b))

    return wrapper

def pear_corr(cr):
    #handle the case where a nan is returned
    if numpy.isnan(cr):
        return 0.0
    if cr < 0.0:
        return 0.0
    else:
        return cr

    #so far in looking cr has never been negative and the scores mostly just sit in the 0.8 to 0.99999 range
    #I am not even sure if cr could be negative with chromatography (indicating inverse relationship between simulation and experiment)
    #return (1.0+cr)/2.0

def cut_zero(times, values, min_value, max_value):
    "cut the raw times and values as close to min_value and max_value as possible and set the rest to zero without smoothing"
    data_zero = numpy.zeros(len(times))
    
    offset = numpy.array([-1, 0, 1])
    
    #find the starting point for the min_index and max_index, the real value could be off by 1 in either direction
    peak_max_index = numpy.argmax(values)
    
    max_index = numpy.argmax(values[:peak_max_index] >= max_value)

    if not max_index:
        #if max_index is zero the whole array is below the value we are searching for so just returnt the whole array
        return values, None, None, None, None

    min_index = numpy.argmax(values[:max_index] >= min_value)
        
    check_max_index = max_index + offset
    check_max_value = values[check_max_index]
    check_max_diff = numpy.abs(check_max_value - max_value)    
    
    check_min_index = min_index + offset
    check_min_value = values[check_min_index]
    check_min_diff = numpy.abs(check_min_value - min_value)
    
    min_index = min_index + offset[numpy.argmin(check_min_diff)]
    max_index = max_index + offset[numpy.argmin(check_max_diff)]

    min_time = times[min_index]
    min_value = values[min_index]

    max_time = times[max_index]
    max_value = values[max_index]
    
    data_zero[min_index:max_index+1] = values[min_index:max_index+1]
    
    return data_zero, min_time, min_value, max_time, max_value

def find_cuts(times, values, spline, spline_der):
    max_index = numpy.argmax(values)
    max_time = times[max_index]
    
    def goal(time):
        return -spline_der(time)
    
    result = scipy.optimize.minimize(goal, max_time, method='powell')
    
    max_time = float(result.x)
    max_value = float(spline(max_time))

    min_index = numpy.argmax(values >= 1e-3*max_value)
    min_time = times[min_index]
    
    def goal(time):
        return abs(spline(time)-1e-3*max_value)
    
    result = scipy.optimize.minimize(goal, min_time, method='powell')
    
    min_time = float(result.x)
    min_value = float(spline(min_time))

    return min_time, min_value, max_time, max_value


# -*- coding: utf-8 -*-

import numpy as np

from emcee.moves.red_blue import RedBlueMove

__all__ = ["DESnookerMove"]


class DESnookerMove(RedBlueMove):
    """A snooker proposal using differential evolution.

    Based on `Ter Braak & Vrugt (2008)
    <http://link.springer.com/article/10.1007/s11222-008-9104-9>`_.

    Credit goes to GitHub user `mdanthony17 <https://github.com/mdanthony17>`_
    for proposing this as an addition to the original emcee package.

    Args:
        gammas (Optional[float]): The mean stretch factor for the proposal
            vector. By default, it is :math:`1.7` as recommended by the
            reference.

    """

    def __init__(self, gammas=1.7, **kwargs):
        self.gammas = gammas
        kwargs["nsplits"] = 4
        super(DESnookerMove, self).__init__(**kwargs)

    def get_proposal(self, s, c, random):
        Ns = len(s)
        Nc = list(map(len, c))
        ndim = s.shape[1]
        q = np.empty((Ns, ndim), dtype=np.float64)
        metropolis = np.empty(Ns, dtype=np.float64)
        for i in range(Ns):
            w = np.array([c[j][random.randint(Nc[j])] for j in range(3)])
            random.shuffle(w)
            z, z1, z2 = w
            delta = s[i] - z
            norm = np.linalg.norm(delta)
            u = delta / np.sqrt(norm)
            q[i] = s[i] + u * self.gammas * (np.dot(u, z1) - np.dot(u, z2))
            q[i] = q[i] % 1
            metropolis[i] = np.log(np.linalg.norm(q[i] - z)) - np.log(norm)
        return q, 0.5 * (ndim - 1.0) * metropolis



import CADETMatch.util as util
import numpy

name = "log"
count = 1
count_extended = 1

def getUnit(location):
    return location.split('/')[3]

def transform(parameter):
    minValue = parameter['min']
    maxValue = parameter['max']

    def trans(i):
        return numpy.log(i)

    return [trans,]

def untransform(seq, cache, parameter):
    values = [numpy.exp(seq[0]),]
    headerValues = values
    return values, headerValues

def untransform_matrix(matrix, cache, parameter):
    values = numpy.exp(matrix)
    return values

untransform_matrix_inputorder = untransform_matrix

def setSimulation(sim, parameter, seq, cache, experiment):
    values, headerValues = untransform(seq, cache, parameter)

    if parameter.get('experiments', None) is None or experiment['name'] in parameter['experiments']:
        location = parameter['location']
    
        try:
            comp = parameter['component']
            bound = parameter['bound']
            index = None
        except KeyError:
            index = parameter['index']
            bound = None

        if bound is not None:
            unit = getUnit(location)
            boundOffset = util.getBoundOffset(sim.root.input.model[unit])

            if comp == -1:
                position = ()
                sim[location.lower()] = values[0]
            else:
                position = boundOffset[comp] + bound
                sim[location.lower()][position] = values[0]

        if index is not None:
            sim[location.lower()][index] = values[0]
    return values, headerValues

def setupTarget(parameter):
    location = parameter['location']
    bound = parameter['bound']
    comp = parameter['component']

    name = location.rsplit('/', 1)[-1]
    sensitivityOk = 1

    try:
        unit = int(location.split('/')[3].replace('unit_', ''))
    except ValueError:
        unit = ''
        sensitivityOk = 0

    return [(name, unit, comp, bound),], sensitivityOk

def getBounds(parameter):
    minValue = numpy.log(parameter['min'])
    maxValue = numpy.log(parameter['max'])

    return [minValue,], [maxValue,]

def getHeaders(parameter):
    location = parameter['location']

    try:
        comp = parameter['component']
    except KeyError:
        comp = 'None'

    name = location.rsplit('/', 1)[-1]
    bound = parameter.get('bound', None)
    index = parameter.get('index', None)
    
    headers = []
    if bound is not None:
        headers.append("%s Comp:%s Bound:%s" % (name, comp, bound))
    if index is not None:
        headers.append("%s Comp:%s Index:%s" % (name, comp, index))
    return headers

def getHeadersActual(parameter):
    return getHeaders(parameter)

def setBounds(parameter, lb, ub):
    parameter['min'] = lb[0]
    parameter['max'] = ub[0]



import CADETMatch.match
import IPython.display as dp
from CADETMatch.cache import Cache
import subprocess
import sys
import pandas
import psutil
import signal

class Match:
    def __init__(self, json_path):
        self.json_path = json_path
        self.cache = Cache()
        self.cache.setup(json_path)

    def start_sim(self):
        #ncpus = psutil.cpu_count(logical=False)
        command = [sys.executable, #'-n', str(ncpus),
                   CADETMatch.match.__file__, str(self.json_path), str(1)]
        pipe = subprocess.PIPE

        proc = subprocess.Popen(command, stdout=pipe, stderr=subprocess.STDOUT, bufsize=1)

        def signal_handler(sig, frame):
            for child in psutil.Process(proc.pid).children(recursive=True):
                child.kill()
            proc.kill()
            print('Terminating')
            sys.exit(0)

        signal.signal(signal.SIGINT, signal_handler)
    
        for line in iter(proc.stdout.readline, b''):
            print(line.decode('utf-8'))
        return proc.poll()

    def plot_corner(self):
        process_dir = self.cache.settings['resultsDirProgress']

        corner = process_dir / 'sns_corner.png'
        corner_transform = process_dir / 'sns_corner_transform.png'

        if corner.exists():
            print("Corner plot in search space")
            image = dp.Image(filename=corner.as_posix(), embed=True)
            dp.display(image)

        if corner_transform.exists():
            print("Corner plot in original space")
            image = dp.Image(filename=corner_transform.as_posix(), embed=True)
            dp.display(image)

    def plot_best(self):
        meta_dir = self.cache.settings['resultsDirMeta']
        best, score, best_score = self.get_best()
     
        for name, col_names in best.items():
        
            print("Best item %s for meta score(s) %s" % (name, ' , '.join(col_names)))
            dp.display(score[name])
            images = meta_dir.glob('%s_*.png' % name)
            for image in images:
                img = dp.Image(filename=image.as_posix(), embed=True)
                dp.display(img)

    def get_best(self):
        "return the best values for each score"
        meta_dir = self.cache.settings['resultsDirMeta']
        csv = meta_dir / 'results.csv'
        data = pandas.read_csv(csv.as_posix())
    
        best = {}
        score = {}
        best_score = {}
    
        cols = [('Product Root Score', False), ('Min Score', False), ('Mean Score', False), ('SSE', False)]
    
        for col_name, order in cols:
            temp = data.sort_values(by=[col_name,], ascending=order)
            head = temp.head(1)
            name = str(head.Name.iloc[0])
            score[name] = head
            best_score[col_name] = head[self.cache.parameter_headers].values[0]
            if name not in best:
                best[name] = [col_name]
            else:
                best[name].append(col_name)

        return best, score, best_score

    def plot_space(self):
        space_dir = self.cache.settings['resultsDirSpace'] / '2d'

        for image in space_dir.glob('*.png'):
            if '1- ' in image.name or 'SSE' in image.name:
                print(image.name)
                img = dp.Image(filename=image.as_posix(), embed=True)
                dp.display(img)